# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
try:
    from pkg_resources import packaging
    Version = packaging.version.Version  # pragma no cover
except ImportError:  # pragma no cover
    # Seen first on Python >=3.11 but should be about `pkg_resources` version
    from packaging.version import Version

import heptapod


def test_package_version():
    # the Version class accepts only PEP440 compliant version strings
    Version(heptapod.__version__)



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

import json
from hgext.clonebundles import (
    GeneratedBundle,
    read_auto_gen,
)
from mercurial import (
    error,
    scmutil,
)
from mercurial.utils import procutil
from minio import Minio, S3Error
import os
import pytest
import re

from heptapod.testhelpers import (
    RepoWrapper,
)
from heptapod.testhelpers.gitlab import (
    GitLabMirrorFixture,
    GitLabStateMaintainerFixture,
)
from heptapod.testhelpers.git import GitRepo
from hgext import clonebundles

from .utils import common_config

from .. import (
    versions as hpd_versions,
    branch as hpd_branch,
    special_ref,
    keep_around,
    ASYNC_HEPTAPOD_CLONE_BUNDLE_GENERATE,
)

parametrize = pytest.mark.parametrize


@pytest.fixture
def wrapper(tmpdir):
    repo_path = tmpdir.join('repo')
    yield RepoWrapper.init(repo_path, config=common_config(tmpdir))


def test_hpd_unique_successor(wrapper, monkeypatch):
    ctx = wrapper.write_commit('foo', message="default0",
                               return_ctx=True)
    repo_path = wrapper.path
    repo_path.join('foo').write('amend 1')
    wrapper.command('amend', message=b'amend1')
    repo_path.join('foo').write('amend 2')
    wrapper.command('amend', message=b'amend2')

    records = []

    def write(*args, **opts):
        records.append((args, opts))

    wrapper.repo.ui.write = write
    wrapper.command('hpd-unique-successor', rev=ctx.hex())
    out = records[0][0][0]

    succ_ctx = scmutil.revsingle(wrapper.repo, out)
    assert succ_ctx.description() == b'amend2'


def test_hpd_unique_successor_divergence(tmpdir, monkeypatch):
    repo_path = tmpdir.join('repo')
    config = common_config(tmpdir)
    config.setdefault('experimental', {})['evolution.allowdivergence'] = 'yes'
    wrapper = RepoWrapper.init(repo_path, config=config)
    ctx = wrapper.write_commit('foo', message="default0",
                               return_ctx=True)
    repo_path.join('foo').write('amend 1')
    wrapper.command('amend', message=b'amend1')

    # let's create the divergence
    wrapper.update(ctx.hex(), hidden=True)
    repo_path.join('foo').write('amend 2')
    wrapper.command('amend', message=b'amend2')

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('hpd-unique-successor', rev=ctx.hex())
    assert 'divergent' in exc_info.value.args[0]


def test_hpd_ensure_gitlab_branches(wrapper):
    # test almost trivial because all the logic is in HeptapodGitHandler
    wrapper.command('hpd-ensure-gitlab-branches')
    assert hpd_branch.read_gitlab_branches(wrapper.repo) == {}


def test_hpd_ensure_gitlab_default_branch(wrapper):
    # initially we don't have any default_gitlab_branch file
    assert not hpd_branch.get_default_gitlab_branch(wrapper.repo)
    # run command to create the file
    wrapper.command('hpd-ensure-gitlab-default-branch')
    assert hpd_branch.get_default_gitlab_branch(wrapper.repo) == b'master'
    # if already exists, running again shouldn't make any difference
    wrapper.command('hpd-ensure-gitlab-default-branch')
    assert hpd_branch.get_default_gitlab_branch(wrapper.repo) == b'master'


def test_hpd_ensure_gitlab_tags(wrapper):
    # test almost trivial because all the logic is in HeptapodGitHandler
    wrapper.command('hpd-ensure-gitlab-tags')
    assert hpd_branch.read_gitlab_tags(wrapper.repo) == {}


def test_hpd_ensure_gitlab_special_refs(wrapper):
    # test almost trivial because all the logic is in HeptapodGitHandler
    wrapper.command('hpd-ensure-gitlab-special-refs')
    assert special_ref.special_refs(wrapper.repo) == {}


def test_hpd_ensure_gitlab_keep_arounds(wrapper):
    # test almost trivial because all the logic is in HeptapodGitHandler
    wrapper.command('hpd-ensure-gitlab-keep-arounds')
    assert list(keep_around.iter_keep_arounds(wrapper.repo)) == []


def test_hpd_ensure_all_gitlab_specific_state_files(wrapper):
    # test almost trivial because all the logic is in HeptapodGitHandler
    wrapper.command('hpd-ensure-all-gitlab-specific-state-files')
    assert special_ref.special_refs(wrapper.repo) == {}
    assert hpd_branch.read_gitlab_tags(wrapper.repo) == {}
    assert hpd_branch.read_gitlab_tags(wrapper.repo) == {}
    assert list(keep_around.iter_keep_arounds(wrapper.repo)) == []
    assert hpd_branch.get_default_gitlab_branch(wrapper.repo) == b'master'


def test_hpd_unique_successor_missing_rev(wrapper, monkeypatch):
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('hpd-unique-successor')
    assert b'specify a revision' in exc_info.value.args[0]


def test_hpd_versions_with_hg_git(tmpdir, monkeypatch):
    # using RepoWrapper is pure lazyness on our part: they  give us the easiest
    # access to fully set up `ui` objects, with activated extensions
    config = common_config(tmpdir)
    config['extensions']['hggit'] = ''
    ui = RepoWrapper.init(tmpdir, config=config).repo.ui
    records = []

    def write(*args, **opts):
        assert all(isinstance(a, bytes) for a in args)
        records.append((args, opts))

    monkeypatch.setattr(ui, 'write', write)
    hpd_versions(ui)
    out = json.loads(records[0][0][0].decode())
    assert set(out.keys()) == {'python', 'mercurial',
                               'topic', 'hggit', 'evolve'}
    # for hggit it looks like: x.y.z (dulwich a.b.c)
    # for Mercurial, it can be just x.y
    version_re = re.compile(r'\d+[.]\d+([.]\d+)?')
    assert all(v is None or version_re.match(v) is not None
               for v in out.values())
    out.pop('hggit', None)  # hggit won't be shipped in some future
    assert all(v is not None for v in out.values())


@pytest.fixture()
def git_mirror_fixture(tmpdir, monkeypatch):
    with GitLabMirrorFixture.init(tmpdir, monkeypatch,
                                  hg_config=common_config(tmpdir)) as fixture:
        yield fixture


def test_git_resync_existing_git_repo(git_mirror_fixture):
    fixture = git_mirror_fixture
    wrapper = fixture.hg_repo_wrapper

    base_ctx = wrapper.commit_file('foo', message='Commit 0')

    # Adding various repo content
    wrapper.command('tag', b'v1.2.3', rev=base_ctx.hex())
    wrapper.commit_file('foo', message='Commit 1')
    # special refs and keep-arounds are added later in this test
    wrapper.command('gitlab-mirror')
    git_repo = fixture.git_repo

    wrapper.set_config('heptapod', 'native', 'yes')

    def assert_initial_git_state():
        assert git_repo.branch_titles() == {b'branch/default': b'Commit 1'}
        assert git_repo.tags() == {b'v1.2.3'}
        # checking ref targets, we don't care about the hashes
        base_git_sha, base_git_title = git_repo.commit_hash_title('v1.2.3')
        assert base_git_title == b'Commit 0'
        return base_git_sha

    # first call does not break anything, and moves Git repo to the dedicated
    # location
    wrapper.command('hpd-git-resync')
    fixture.reload_git_repo()
    assert fixture.git_repo.path != git_repo.path
    git_repo = fixture.git_repo
    assert_initial_git_state()

    # now adding a new commit to check basic incrementality
    # (adding another branch just to spice it a bit)
    wrapper.commit_file('foo', message='Commit 2')
    wrapper.commit_file('bar', parent=base_ctx, branch='other',
                        message='Commit 3')
    wrapper.command('gitlab-mirror')

    # Mirroring to Git did not happen (validity of test hypothesis)
    assert git_repo.branch_titles() == {b'branch/default': b'Commit 1'}

    wrapper.command('hpd-git-resync')
    assert git_repo.branch_titles() == {b'branch/default': b'Commit 2',
                                        b'branch/other': b'Commit 3'}
    assert git_repo.tags() == {b'v1.2.3'}


@pytest.fixture()
def native_fixture(tmpdir, monkeypatch):
    with GitLabStateMaintainerFixture.init(
            tmpdir, monkeypatch,
            hg_config=common_config(tmpdir)) as fixture:
        fixture.hg_repo_wrapper.set_config('heptapod.native', True)
        yield fixture


def test_git_resync_creates_git_repo(native_fixture):
    fixture = native_fixture
    wrapper = fixture.hg_repo_wrapper

    base_ctx = wrapper.commit_file('foo', message='Commit 0')

    # Adding various repo content
    wrapper.command('tag', b'v1.2.3', rev=base_ctx.hex())
    wrapper.commit_file('foo', message='Commit 1')
    # TODO special ref
    # TODO keep around
    wrapper.command('gitlab-mirror')
    wrapper.command('hpd-export-native-to-git')
    git_repo = GitRepo(
        fixture.base_path / '+hgitaly/hg-git'
        / wrapper.path.basename.replace('.hg', '.git')
    )
    assert git_repo.branch_titles() == {b'branch/default': b'Commit 1'}
    assert git_repo.tags() == {b'v1.2.3'}
    # checking ref targets, we don't care about the hashes
    base_git_sha, base_git_title = git_repo.commit_hash_title('v1.2.3')
    assert base_git_title == b'Commit 0'
    return base_git_sha


def test_move_out_of_gitaly_reach(git_mirror_fixture):
    fixture = git_mirror_fixture
    wrapper = fixture.hg_repo_wrapper
    orig_path = fixture.git_repo.path

    with pytest.raises(error.Abort):
        wrapper.command('move-hg-git-repo-out-of-gitaly-reach')

    wrapper.set_config('heptapod', 'native', True)
    wrapper.command('move-hg-git-repo-out-of-gitaly-reach')

    fixture.reload_git_repo()
    git_repo = fixture.git_repo
    assert git_repo.path != orig_path
    assert not os.path.exists(orig_path)
    assert os.path.exists(git_repo.path)
    # would not work if it were not a Git repo
    assert git_repo.branches() == {}


class CloneBundlesFixture():
    def __init__(self, parent, monkeypatch):
        self.native_fixture = parent
        wrapper = self.hg_repo_wrapper = parent.hg_repo_wrapper

        self.bucket = 'seau'
        self.bucket_subdir = 'group/subgrp/project-123'
        self.pub_base_url = 'https://pub.hgcb.test'

        self.set_conf('trigger.below-bundled-ratio', 1.0)
        self.set_conf('auto-generate.formats', 'zstd-v2')
        self.set_conf('clone-bundles.public-base-url', self.pub_base_url,
                      section='heptapod')
        self.set_conf('clone-bundles.s3.endpoint', 's3.test',
                      section='heptapod')
        self.set_conf('clone-bundles.s3.public-bucket', self.bucket,
                      section='heptapod')

        wrapper.repo.ui.environ.update((
            (b'GL_REPOSITORY', b'project-123'),
            (b'HEPTAPOD_HGRC', b'/some/heptapod.hgrc'),
        ))

        self.fput_records = []
        self.rm_records = []

        monkeypatch.setattr(
            Minio, 'fput_object',
            lambda client, *a, **kw: self.fput_records.append((a, kw))
        )
        monkeypatch.setattr(
            Minio, 'remove_object',
            lambda client, *a, **kw: self.rm_records.append((a, kw))
        )

    def set_conf(self, k, v, section='clone-bundles'):
        self.hg_repo_wrapper.repo.ui.setconfig(
            section.encode(), k.encode(), str(v).encode())


@pytest.fixture
def clone_bundles_fixture(native_fixture, monkeypatch):
    yield CloneBundlesFixture(native_fixture, monkeypatch)


# with 'disabled', the point is to run everything configured
# as for 'command' and check that is really does nothing
@parametrize('trigger', ('command', 'on-change', 'disabled'))
def test_clone_bundles_refresh(clone_bundles_fixture, monkeypatch, trigger):
    fixture = clone_bundles_fixture
    wrapper = fixture.hg_repo_wrapper
    on_change = trigger == 'on-change'

    if on_change:
        fixture.set_conf('auto-generate.on-change', True)
        wrapper.repo.ui.environ.update((
            (b'HEPTAPOD_PUBLIC_REPOSITORY', b'true'),
            (b'HEPTAPOD_PROJECT_NAMESPACE_FULL_PATH', b'group/subgrp'),
        ))
        bg_commands = []
        monkeypatch.setattr(procutil, 'runbgcommand',
                            lambda *a, **kw: bg_commands.append((a, kw)))

    fput_records = fixture.fput_records

    if trigger == 'disabled':
        wrapper.repo.ui.setconfig(b'heptapod', b'clone-bundles', b'disabled')

    base_ctx = wrapper.commit_file('foo', message='Commit 0')

    # Adding various repo content
    wrapper.command('tag', b'v1.2.3', rev=base_ctx.hex())
    wrapper.commit_file('foo', message='Commit 1')

    if on_change:
        assert len(bg_commands) == 3  # indeed, we had 3 transactions
        (cmd, env), _ = bg_commands[-1]
        assert env[b'GL_REPOSITORY'] == b'project-123'
        assert env[b'HGRCPATH'] == b'/some/heptapod.hgrc'
        assert env.get(b'HEPTAPOD_HG_BUNDLE_SPEC')
        # let us launch it synchronously instead and benefit
        # from our Minio monkey patch
        wrapper.repo.ui.environ.update(env)
        wrapper.command(ASYNC_HEPTAPOD_CLONE_BUNDLE_GENERATE)
    else:
        wrapper.command('hpd-clone-bundles-refresh',
                        project_namespace_full_path=b'group/subgrp',
                        public=True,
                        )
        if trigger == 'disabled':
            assert not fput_records
            return

    fput_args = fput_records[0][0]
    assert fput_args[0] == fixture.bucket
    basename = fput_args[2].rsplit('/', 1)[1]
    assert fput_args[1] == '/'.join((fixture.bucket_subdir, basename))

    bundles = read_auto_gen(wrapper.repo)
    assert len(bundles) == 1
    bundle = bundles[0]
    assert isinstance(bundle, GeneratedBundle)
    assert bundle.ready
    assert bundle.basename == basename.encode()
    assert bundle.revs == 3
    assert bundle.file_url.decode() == '/'.join(
        (fixture.pub_base_url, fixture.bucket_subdir, basename)
    )


@parametrize('result', ('success', 's3_failure'))
def test_delete_clone_bundle(clone_bundles_fixture, monkeypatch, result):
    fixture = clone_bundles_fixture
    wrapper = fixture.hg_repo_wrapper
    rm_records = fixture.rm_records

    wrapper.commit_file('foo', message='Commit 0')
    wrapper.command('hpd-clone-bundles-refresh',
                    project_namespace_full_path=b'group/subgrp',
                    public=True,
                    )

    bundles = read_auto_gen(wrapper.repo)
    assert len(bundles) == 1
    bundle = bundles[0]
    assert isinstance(bundle, GeneratedBundle)

    def find_outdated_bundles(repo, bundles):
        return [bundle]

    if result == 's3_failure':
        def rm_obj(*a, **kw):
            raise S3Error(403,
                          message='patched',
                          resource=bundle.basename,
                          request_id='test-req',
                          host_id='some-host',
                          response=403)

        monkeypatch.setattr(Minio, 'remove_object', rm_obj)

    monkeypatch.setattr(clonebundles, 'find_outdated_bundles',
                        find_outdated_bundles)

    clonebundles.collect_garbage(wrapper.repo)

    if result == 'success':
        assert len(rm_records) == 1
        rm_args = rm_records[0][0]
        assert rm_args[0] == fixture.bucket
        basename = bundle.basename.decode('ascii')
        assert rm_args[1] == '/'.join((fixture.bucket_subdir, basename))
    else:
        # no crash and nothing else done
        assert len(rm_records) == 0



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Git test helpers based on the actual Git

This module does not depend on the dulwich library, hence avoiding
a Python dependency and potential tautology problems in the testing
of code that relies on dulwich.

The price to pay for that is to assume the non-Python dependency on Git,
typically solved by installing Git system-wide.
"""
import os
import re
import shutil
import subprocess
import tempfile
from mercurial import pycompat

ZERO_SHA = b'0' * 40
"""Git null commit.

Used notably to represent creation and deletion of Git refs.

It is actually identical to Mercurial's `nullid`, but that could change
at some point.

It is also identical to ``dulwich.protocol.ZERO_SHA``, but we don't want
to depend on dulwich here.
"""


def extract_git_branch_titles(branches):
    return {ref: info['title'] for ref, info in branches.items()}


def git_version():
    out = subprocess.check_output(('git', 'version')).decode().strip()
    m = re.match(r'git version (\d+)\.(\d+)\.(\d+)', out)
    return tuple(int(m.group(i)) for i in range(1, 4))


GIT_VERSION = git_version()


class GitRepo(object):
    """Represent a Git repository, relying on the Git executable.

    The ``git`` command is expected to be available on the ``PATH``.
    """

    def __init__(self, path):
        self.path = path

    @classmethod
    def init(cls, path, bare=True):
        cmd = ['git', 'init']
        if bare:
            cmd.append('--bare')
        cmd.append(str(path))
        subprocess.call(cmd)
        return cls(path)

    @classmethod
    def create_from_bundle_data(cls, target_path, data,
                                tmpdir=None, stream=False):
        """Create a Git repository from a bundle.

        The data is dumped to a temporary file before calling Git.

        :param stream: see ``data``
        :param bundle_data: either :class:`bytes` if ``stream`` is False
          or an iterable of :class: `bytes` if ``stream`` is True
        :param tmpdir: passed over to ``tempfile``.
        """
        with tempfile.NamedTemporaryFile(dir=tmpdir) as bundlef:
            if stream:
                for chunk in data:
                    bundlef.write(chunk)
            else:
                bundlef.write(data)
            bundlef.flush()

            subprocess.check_call(
                ('git', 'clone', '--bare', bundlef.name, str(target_path))
            )
        return cls(target_path)

    def git(self, *args):
        return subprocess.check_output(('git', ) + args, cwd=str(self.path))

    def delete(self):  # pragma no cover (see below)
        # We just removed the unique caller of this method. It was a direct
        # call from a test, not something deeply buried in fixture code or
        # similar (coverage drop would have been suspicious in that case).
        if os.path.exists(self.path):
            shutil.rmtree(self.path)

    def branch_titles(self):
        return extract_git_branch_titles(self.branches())

    def branches(self):
        out = self.git('branch', '-v', '--no-abbrev')
        split_lines = (l.lstrip(b'*').strip().split(None, 2)
                       for l in out.splitlines())
        return {sp[0]: dict(sha=sp[1], title=sp[2]) for sp in split_lines}

    def tags(self):
        out = self.git('tag')
        return set(l.strip() for l in out.splitlines())

    def commit_hash_title(self, revspec):
        out = self.git('log', '-n1', revspec, r'--pretty=format:%H|%s')
        return out.strip().split(b'|')

    def get_symref(self, name):
        dotgit = self.path.join('.git')
        path = dotgit if dotgit.exists() else self.path
        return path.join(name).read().strip().split(':', 1)[1].strip()

    def set_symref(self, name, target_ref):
        self.path.join(name).write('ref: %s\n' % target_ref)

    def set_branch(self, name, sha):
        sha = pycompat.sysstr(sha)
        self.path.join('refs', 'heads', name).ensure().write(sha + '\n')

    def get_branch_sha(self, name):
        if isinstance(name, bytes):
            name = name.decode()
        return self.path.join('refs', 'heads', name).read().strip()

    def write_ref(self, ref_path, value):
        """Call git to write a ref.

        symbolic refs are taken at face value, hence
        ``write_git_ref(repo, b'HEAD', b'refs/heads/mybranch')`` will move
        ``HEAD`` to ``mybranch``, instead of moving whatever ``HEAD``
        points to.
        """
        self.git('update-ref', b'--no-deref', ref_path, value)

    def all_refs(self):
        """Call git to return all refs.

        :return: a :class:`dict` whose keys are full ref paths and values are
           Git commit SHAs, both as :class:`bytes`.
        """
        lines = self.git('show-ref').splitlines()
        return {split[1]: split[0]
                for split in (line.split(b' ', 1) for line in lines)}

    def raw_refs(self, with_root_refs=False):
        """Use `git for-each-ref` to query all refs, including "root refs"

        The output is the full `stdout` of the Git process, as bytes.
        Each line has 2 or 3 fields:

        - the target object SHA
        - the name of the ref (e.g, `refs/heads/main` or `HEAD`)
        - its target ref *if* it is a root ref (e.g, `HEAD`).

        The root refs are a subset of symrefs. It includes `HEAD` but not
        arbitrary symrefs that can be created with `git symbolic-ref`.

        Git sorts the output on `refname` (e.g., `refs/heads/main` or `HEAD`)
        by default, which is fine to compare two repositories.
        """
        git_args = ['for-each-ref',
                    '--format', '%(objectname)s %(refname) %(symref)']
        if with_root_refs:
            git_args.append('--include-root-refs')  # pragma no cover

        return self.git(*git_args)

    def __eq__(self, other):
        """True iff the two repos have the same content.

        THis is done by comparing all refs, including the remotes
        """
        # not insisting on the same exact class
        if not isinstance(other, GitRepo):
            return False

        opts = dict(with_root_refs=True)
        if GIT_VERSION < (2, 45, 0):  # pragma no cover
            opts['with_root_refs'] = False
            if self.get_symref('HEAD') != other.get_symref('HEAD'):
                return False

        return self.raw_refs(**opts) == other.raw_refs(**opts)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Interaction with Git repos.

It is not expected that the rise of the HGitaly project would remove
all interesting things to do with Git.
"""
from __future__ import absolute_import

from mercurial import error
try:
    from hgext3rd.hggit.git_handler import GitHandler
except ImportError:  # pragma: no cover (fallback for hg-git < 0.11)
    from hggit.git_handler import GitHandler

from heptapod.gitlab.branch import (
    GITLAB_BRANCH_REF_PREFIX,
    gitlab_branch_ref as git_branch_ref,
    gitlab_branch_from_ref as git_branch_from_ref,
)
from heptapod.gitlab.tag import (
    GITLAB_TAG_REF_PREFIX,
)
from .typed_ref import (
    GITLAB_TYPED_REFS_MISSING,
    gitlab_typed_refs,
)
from mercurial import (
    hg,
)
import os
import re

from .branch import (
    get_default_gitlab_branch,
    set_default_gitlab_branch,
    gitlab_branches,
    gitlab_tags,
    write_gitlab_branches,
    write_gitlab_tags,
)
from .special_ref import (
    special_refs,
    write_special_refs
)
from .keep_around import (
    iter_keep_arounds,
    init_keep_arounds,
)
from .state_maintainer import GitLabStateMaintainer


class HeptapodGitHandler(GitHandler, GitLabStateMaintainer):

    def __init__(self, *args, **kwargs):
        super(HeptapodGitHandler, self).__init__(*args, **kwargs)

        main_repo = hg.sharedreposource(self.repo)
        if main_repo is None:
            main_repo = self.repo

        self.gitdir = re.sub(br'\.hg$', b'', main_repo.root) + b'.git'
        self.unfiltered_repo = self.repo.unfiltered()
        self._default_git_ref = None

    def native_project_ensure_set_git_dir(self):
        """Ensure existence of the auxiliary git repo for native Projects.

        Even native Mercurial projects can have auxiliary Git repositories,
        for mirroring to Git, although they are not alongside the Mercurial
        repository and are managed in a different way (see heptapod#1963).

        This makes sure that any Git repository for a native project is moved
        over to the proper location.
        """
        storages_root = self.repo.ui.config(b'heptapod', b'repositories-root')
        git_repos_root = os.path.join(storages_root,
                                      b'+hgitaly', b'hg-git')
        os.makedirs(git_repos_root, mode=0o700, exist_ok=True)

        # After instantiation, `self.gitdir` is the proper one for
        # legacy (hg-git based) projects
        old_gitdir = self.gitdir

        rel_path = os.path.relpath(old_gitdir, storages_root)
        self.gitdir = os.path.join(git_repos_root, rel_path)

        os.makedirs(os.path.dirname(self.gitdir),
                    mode=0o700,
                    exist_ok=True)

        if os.path.exists(old_gitdir) and not os.path.exists(self.gitdir):
            os.rename(old_gitdir, self.gitdir)

        self.fixup_git_pool()

    def fixup_git_pool(self):
        alternates_path = os.path.join(self.gitdir,
                                       b'objects/info/alternates')
        if not os.path.exists(alternates_path):
            return

        with open(alternates_path, 'rb') as alternates:
            lines = alternates.readlines()

        new_lines = []
        for line in lines:
            line = line.strip()
            tgt_path = os.path.join(self.gitdir, line)
            if os.path.exists(tgt_path):
                new_lines.append(line)
            else:
                repositories_root = self.repo.ui.config(b'heptapod',
                                                        b'repositories-root')
                idx = tgt_path.find(b'@pools/')
                if idx == -1:
                    raise error.Abort(
                        b"Unreachable Git alternates path that is not "
                        b"under the expected GitLab pools: %r" % tgt_path)
                abs_tgt = os.path.join(repositories_root, tgt_path[idx:])
                if not os.path.exists(abs_tgt):
                    raise error.Abort(
                        b"After fix, path to shared Git inside GitLab pools "
                        b"does not exist: %r" % abs_tgt)
                new_lines.append(os.path.relpath(abs_tgt,
                                                 self.gitdir + b'/objects'))
        with open(alternates_path, 'wb') as alternates:
            alternates.write(b'\n'.join(new_lines))

    @property
    def gitlab_refs(self):
        return self.git.refs

    def has_git_repo(self):
        return os.path.exists(self.gitdir)

    def get_default_gitlab_ref(self):
        # HEAD is the source of truth for GitLab default branch
        # For native projects, it's going to be the one stored
        # in Mercurial, returned by get_default_gitlab_branch(), and we are
        # maintaining consistency until we reach the HGitaly2 milestone
        # (i.e., as long as hg-git is used even for native projects)
        res = self._default_git_ref
        if res is None:
            res = self.git.refs.get_symrefs().get(b'HEAD')
            self._default_git_ref = res
        return res

    def ensure_gitlab_default_branch(self):
        """Init GitLab default branch file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab default branch :class:`bytes`.
        """
        repo = self.repo
        gl_branch = get_default_gitlab_branch(repo)
        if gl_branch is not None:
            return

        self.ui.note(b"Initializing GitLab default branch file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            gl_branch_ref = self.get_default_gitlab_ref()
            gl_branch = git_branch_from_ref(gl_branch_ref)
        else:
            gl_branch = b'master'

        set_default_gitlab_branch(repo, gl_branch)
        return gl_branch

    def set_default_gitlab_ref(self, new_default_ref):
        new_gl_branch = git_branch_from_ref(new_default_ref)
        self.repo.ui.note(
            b"Setting Git HEAD to %s and Hg default "
            b"GitLab branch to %s" % (new_default_ref, new_gl_branch))
        self.git.refs.set_symbolic_ref(b'HEAD', new_default_ref)
        set_default_gitlab_branch(self.repo, new_gl_branch)

        # cache invalidation
        self._default_git_ref = None

    hg_sha_from_gitlab_sha = GitHandler.map_hg_get
    gitlab_sha_from_hg_sha = GitHandler.map_git_get

    def extract_all_gitlab_refs(self):
        """Heptapod version of GitHandler.get_exportable().

        This rewraps :meth:`GitHandler.get_exportable` to add named branches
        and topics to the returned Git refs
        """
        git_refs = super(HeptapodGitHandler, self).get_exportable()
        self.extract_current_gitlab_branches(git_refs)
        return git_refs

    def export_commits(self):
        try:
            self.export_git_objects()
            self.update_gitlab_references()
        finally:
            self.save_map(self.map_file)

    def ensure_gitlab_branches(self):
        """Init GitLab branches state file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab branches :class:`dict`, same as
           ``gitlab_branches(self.repo)`` would.
        """
        repo = self.repo
        if gitlab_branches(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab branches state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            gl_branches = self.gitlab_branches_hg_shas(self.git.refs.as_dict())
        else:
            gl_branches = {}
        write_gitlab_branches(repo, gl_branches)
        return gl_branches

    def ensure_gitlab_tags(self):
        """Init GitLab tags state file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab tags :class:`dict`, same as
           ``gitlab_tags(self.repo)`` would.
        """
        repo = self.repo
        if gitlab_tags(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab tags state file "
                     b"for repo at '%s'" % repo.root)
        if self.has_git_repo():
            gl_tags = self.gitlab_tags_hg_shas(self.git.refs)
        else:
            gl_tags = {}
        write_gitlab_tags(repo, gl_tags)
        return gl_tags

    def ensure_gitlab_special_refs(self):
        """Init GitLab special refs file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab special refs :class:`dict`, same as
           ``special_refs(self.repo)`` would.
        """
        repo = self.repo
        if special_refs(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab special refs state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            srefs = self.gitlab_special_refs_hg_shas(self.git.refs)
        else:
            srefs = {}
        write_special_refs(repo, srefs)
        return srefs

    def ensure_gitlab_keep_arounds(self):
        """Init GitLab keep arounds file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab keep-arounds :class:`set`.
        """
        repo = self.repo
        try:
            if next(iter_keep_arounds(repo)) is not GITLAB_TYPED_REFS_MISSING:
                return
        except StopIteration:
            return

        self.ui.note(b"Initializing GitLab keep-arounds state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            kas = self.gitlab_keep_arounds(self.git.refs)
        else:
            kas = ()
        init_keep_arounds(repo, kas)
        return kas

    def force_git_refs_from_gitlab_files(self):
        repo = self.repo
        git_refs = self.git.refs
        refs_to_remove = git_refs.keys()
        git_sha_from = self.gitlab_sha_from_hg_sha

        for type_name, ref_prefix in (
                ('branches', GITLAB_BRANCH_REF_PREFIX),
                ('tags', GITLAB_TAG_REF_PREFIX),
        ):
            typed_refs = gitlab_typed_refs(repo, type_name)
            if typed_refs is GITLAB_TYPED_REFS_MISSING:
                continue
            for name, hg_sha in typed_refs.items():
                ref = ref_prefix + name
                refs_to_remove.discard(ref)

                git_sha = git_sha_from(hg_sha)
                if git_sha is None:
                    self.repo.ui.warn(
                        b"force_git_refs_from_gitlab_files: did not find "
                        b"Git SHA for Mercurial %r (ref %r)" % (hg_sha, ref)
                    )
                else:
                    git_refs[ref] = git_sha

        for ref in refs_to_remove:
            del git_refs[ref]

        default_gl_branch = get_default_gitlab_branch(self.repo)
        if default_gl_branch is not None:
            # don't use `self.set_default_gitlab_ref` as it writes also
            # the state file in the hg repo (should be idempotent, but let's
            # not risk that)
            self.git.refs.set_symbolic_ref(
                b'HEAD',
                git_branch_ref(default_gl_branch))



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import subprocess
from ..git import GitRepo
import pytest

parametrize = pytest.mark.parametrize


def test_refs(tmpdir):
    server = tmpdir / 'server.git'
    # server_repo is bare, hence we cannot commit directly in there
    server_repo = GitRepo.init(server)

    client = tmpdir / 'client'
    client_repo = GitRepo.init(client, bare=False)
    client_repo.git('remote', 'add', 'origin', str(server))
    client_repo.git('config', 'user.name', 'John Doe')
    client_repo.git('config', 'user.email', 'jdoe@heptapod.example')

    (client / 'foo').write('foo')
    client_repo.git('add', 'foo')
    client_repo.git('commit', '-m', 'init commit', '--no-gpg-sign')
    client_repo.git('push', 'origin', 'master')

    assert server_repo.branch_titles() == {b'master': b'init commit'}

    # correct commit SHA is read from server repo
    sha = server_repo.get_branch_sha('master')
    client_log = subprocess.check_output(['git', 'log', '--oneline', sha],
                                         cwd=str(client))
    assert client_log.split(b' ', 1)[-1].strip() == b'init commit'

    # Testing more commit and branch methods
    assert server_repo.commit_hash_title('master') == [sha.encode(),
                                                       b'init commit']
    assert server_repo.get_branch_sha(b'master') == sha

    client_repo = GitRepo(client)
    assert server_repo != client  # wrong class
    assert server_repo != client_repo  # the client has remote refs
    client_repo.git('update-ref', '-d', 'refs/remotes/origin/master')
    assert server_repo == client_repo

    # (for Git < 2.45 coverage): same HEAD but an additional branch
    (client / 'foo').write('bar')
    client_repo.git('checkout', '-b', 'other')
    client_repo.git('add', 'foo')
    client_repo.git('commit', '-m', 'other branch', '--no-gpg-sign')
    assert server_repo != client_repo


@parametrize('stream', ('stream', 'bytes'))
def test_create_from_bundle_data(stream, tmpdir):
    src_path = tmpdir / 'src'
    src = GitRepo.init(src_path, bare=False)
    (src_path / 'foo').write('foo')
    src.git('checkout', '-b', 'main')
    src.git('add', 'foo')
    src.git('config', 'user.name', 'John Doe')
    src.git('config', 'user.email', 'jdoe@heptapod.example')
    src.git('commit', '-m', "commit to bundle", '--no-gpg-sign')

    bundle_data = src.git('bundle', 'create', '-q', '-', '--all')
    stream = stream == 'stream'
    if stream:
        bundle_data = [bundle_data[:10], bundle_data[10:]]

    target_repo = GitRepo.create_from_bundle_data(
        target_path=tmpdir / 'target',
        data=bundle_data,
        stream=stream
    )
    assert target_repo.branch_titles() == {b'main': b"commit to bundle"}
